from .library import DataIntegrityError
from .library import PeerDoesExistError
from .library import PeerDoesNotExistError
from .library import SettingDoesNotExistError
from .library import Site
from .library import SiteDoesExistError
from .library import SiteDoesNotExistError
from .library import WireguardNotFoundError
from .library import WireUI

from .ui import run_ui

__all__ = [
    "DataIntegrityError",
    "PeerDoesExistError",
    "PeerDoesNotExistError",
    "SettingDoesNotExistError",
    "Site",
    "SiteDoesExistError",
    "SiteDoesNotExistError",
    "WireguardNotFoundError",
    "WireUI",
]



# console.py
# Write text and error messages to console
# Author: Tim Schlottmann

import sys

_verbosity = 2


def set_verbosity(verbosity: int = 0):
  """Sets the verbosity level of the console output"""

  global _verbosity

  assert isinstance(
      verbosity, int
  ), f"verbosity should be <class \'int\'> and not {str(type(verbosity))}"
  assert verbosity >= 0, f"verbosity should be >= 0 and not {verbosity}"
  _verbosity = verbosity


def print_message(verbosity: int, *args, **kwargs):
  """Prints a message to STDOUT

  The message is only printed, if verbosity is greater than the setted verbosity
  """

  if verbosity <= _verbosity:
    print(*args, file=sys.stdout, **kwargs)


def print_error(verbosity: int, *args, **kwargs):
  """Prints a message to STDERR

  The message is only printed, if verbosity is greater than the setted verbosity
  """

  if verbosity <= _verbosity:
    print(*args, file=sys.stderr, **kwargs)



class Error(Exception):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)


class KeyExistenceError(Error):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)


class KeyDoesExistError(KeyExistenceError):
  def __init__(self, key_name):
    super().__init__(f"{key_name} does already exist")


class KeyDoesNotExistError(KeyExistenceError):
  def __init__(self, key_name):
    super().__init__(f"{key_name} does not exist")


class PeerDoesExistError(KeyDoesExistError):
  pass


class PeerDoesNotExistError(KeyDoesNotExistError):
  pass


class SettingDoesExistError(KeyDoesExistError):
  pass


class SettingDoesNotExistError(KeyDoesNotExistError):
  pass


class SiteDoesExistError(KeyDoesExistError):
  pass


class SiteDoesNotExistError(KeyDoesNotExistError):
  pass


class WireguardNotFoundError(Error):
  pass


class DataIntegrityError(Error):
  pass



# sites.py
# Wireguard sites
# Author: Tim Schlottmann

from json import dumps

from typing import NamedTuple
from typing import Union

from .dicts import JsonDict
from .dicts import JsonDict
from .peers import PeerItems
from .peers import Peers


#Evaluate TypedDict when python3.8 is available
SiteItems = dict


class Sites(JsonDict):
  """ Sites for wireguard """
  def __init__(self,
               initialdata: Union[dict, str] = {},
               defaults: Union[dict, str] = {}):

    super().__init__(initialdata, defaults)

  def __getitem__(self, site_name) -> SiteItems:
    return super().__getitem__(site_name)

  def __setitem__(self, site_name, site: SiteItems):
    super().__setitem__(site_name, site)



import io
import json
import os
import subprocess
import tempfile

from .typedefs import ConnectionTable
from .typedefs import JsonDict
from .typedefs import JSONDecodeError


def write_file(path: str, s: str = "") -> str:
  """ Save a string to a file """
  with open(path, "w") as f:
    f.write(s)
    f.flush()
  return path


def read_file(path: str) -> str:
  """ Get the content of a file """
  try:
    with open(path, "r") as f:
      return f.read()
  except FileNotFoundError:
    return ""


# def write_jsonfile(path: str, d: JsonDict):
#   write_file(path, str(d))

# def read_jsonfile(path: str) -> JsonDict:
#   try:
#     return JsonDict(read_file)
#   except JSONDecodeError as e:
#     raise e

def edit_string(s: str = "") -> str:
  with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as f:
    n = f.name
    f.write(s)

  subprocess.run(["editor", n])
  s = read_file(n)

  os.remove(n)

  return s

def edit_dict(d: JsonDict = JsonDict()) -> JsonDict:
  valid = False
  while not valid:
    try:
      d = JsonDict(edit_string(str(d)))
    except json.JSONDecodeError:
      pass
    else:
      valid = True
  return d

def edit_connection_table(ct: ConnectionTable) -> ConnectionTable:
  s = ""
  valid = False
  while not valid:
    try:
      ct.update(edit_string(str(ct) + s))
    except ValueError as e:
      s = f"\n{e}"
    else:
      valid = True
  return ct

def prepare_directory(path: str):
  """ Clears existing config files in a directory """

  if not os.path.isdir(path):
    os.makedirs(path)
  else:
    _clean_directory(path)


def _clean_directory(path: str):
  files = os.listdir(path)
  for f in files:
    if os.path.isfile(os.path.join(path, f)):
      os.remove(os.path.join(path, f))


def delete_directory(path: str):
  """ Clears existing config files in a directory """

  if os.path.isdir(path):
    _clean_directory(path)
    os.rmdir(path)



import ipaddress

from .console import print_error, print_message

from ..library import ConnectionTable
from ..library import Peer
from ..library import WireUI

def yes_no_menu(string) -> bool:
  valid = False
  while not valid:
    choice = input(string + " [y/n] ")
    if choice == "y":
      choice = True
      valid = True
    elif choice == "n":
      choice = False
      valid = True
  return choice

def create_wireguard_config(w: WireUI, site_name: str):
  created_files = w.create_wireguard_config(site_name)
  print_message(1, "The following files have been created:")
  for f in created_files:
    print_message(1, f)
  print_message(0, f"{len(created_files)} file(s) have been created.")


def get_new_peer_properties(w: WireUI, site_name: str, peer_name: str, ct: ConnectionTable) -> Peer:
  print_message(0, f"Collecting properties of peer {peer_name}...")

  # If peer has ingoing connections endpoint and port is needed
  if ct.get_ingoing_connected_peers(peer_name):
    endpoint = get_endpoint()
    port = get_port()
    persistent_keep_alive = get_persistent_keep_alive()
    allow_ipv4 = False
    allow_ipv6 = False
    for n in w.get_networks(site_name):
      v = ipaddress.ip_network(n).version
      if v == 4:
        allow_ipv4 = True
      elif v == 6:
        allow_ipv6 = True
    additional_allowed_ips = get_additional_allowed_ips(allow_ipv4, allow_ipv6)
  else:
    endpoint = ""
    port = 0
    persistent_keep_alive = -1
    additional_allowed_ips = []

  # If peer has outgoing connections redirect_all_traffic is needed
  if ct.get_outgoing_connected_peers(peer_name):
    redirect_all_traffic = get_redirect_all_traffic()
  else:
    redirect_all_traffic = None

  return Peer(peer_name, additional_allowed_ips, ct.get_outgoing_connected_peers(peer_name), ct.get_main_peer(peer_name), ct.get_ingoing_connected_peers(peer_name), endpoint, port, persistent_keep_alive, redirect_all_traffic)


# TODO: is correct messages
def get_endpoint() -> str:
  return input("Please enter the URL or IP address of the server: ")


# TODO: is correct messages
def get_port() -> int:
  while True:
    port = input("Please enter the port the adapter should listen on: ")
    try:
      port = int(port)
    except ValueError:
      print_error(0, "Error: The port was not a valid integer.")
      continue
    else:
      if port <= 0 or port > 65535:
        print_error(0, "Error: The port should be between 0 and 65535")
        continue
    return port


def get_persistent_keep_alive() -> int:
  if yes_no_menu("Is the peer behind a NAT?"):
    return 25
  else:
    return 0


# TODO: is correct messages
def get_additional_allowed_ips(allow_ipv4: bool, allow_ipv6: bool) -> list:
  l = []
  if yes_no_menu("Do you want to add an additional AllowedIP network?"):
    while True:
      while True:
        try:
          a = ipaddress.ip_network(input("Please enter an additional ip network to add to the AllowedIPs List: "))
        except ValueError as e:
          print_error(0, "Error: Input is not a valid IP network.")
          print_error(2, e)
        else:
          if not allow_ipv4 and a.version == 4:
            print_message(0, "IPv6 only network. An IPv4 address is not allowed!")
          elif not allow_ipv6 and a.version == 6:
            print_message(0, "IPv4 only network. An IPv6 address is not allowed!")
          else:
            break

      l.append(str(a.with_prefixlen))

      if yes_no_menu("Do you want to add another network?"):
        continue
      else:
        break

  return l


# TODO: is correct messages
def get_redirect_all_traffic() -> bool:
  return yes_no_menu("Please enter if all traffic from the peer should be redirected:")



# settings.py
# Settings
# Author: Tim Schlottmann

from typing import Any

from .dicts import JsonDict


class Settings(JsonDict):
  """ Settings of the app """

  def __getitem__(self, setting_name) -> Any:
    return super().__getitem__(setting_name)

  def __setitem__(self, setting_name, setting):
    super().__setitem__(setting_name, setting)



# wireguard.py
# Get keys from wg executable
# Author: Tim Schlottmannimport os

import subprocess
import collections

from .typedefs import Keys
from .typedefs import WireguardNotFoundError


def get_keys() -> Keys:
  """ Creates private, public and preshared key """

  privkey = _get_privkey()
  return Keys({
      "privkey": privkey,
      "pubkey": _get_pubkey(privkey),
      "psk": _get_psk(),
  })


def _get_privkey() -> str:
  """ Get a private key from wg """

  return _run(["wg", "genkey"])


def _get_pubkey(private: str) -> str:
  """ Get a public key from wg """

  return _run(["wg", "pubkey"], input=private.encode("utf-8"))


def _get_psk() -> str:
  """ Get a presharedkey from wg """

  return _run(["wg", "genpsk"])


def _run(*args, **kwargs) -> str:
  """ Run a program on the OS and collect output """

  try:
    return subprocess.run(
        *args, stdout=subprocess.PIPE,
        **kwargs).stdout.decode("utf-8").strip()
  except FileNotFoundError as e:
    raise WireguardNotFoundError(
        "Wireguard not found. Please install it and/or make shure it is available via $PATH")



# config.py
# Create and write wireguard config files
# Author: Tim Schlottmann

import ipaddress
import os

from .typedefs import Keys
from .typedefs import PeerItems
from .typedefs import Peers
from .typedefs import SiteItems
from .io_ import delete_directory
from .io_ import prepare_directory
from .io_ import write_file


def write_config(site: SiteItems, wg_config_path: str) -> list:
  """ Create a json config file from the site parameters """

  peer_addresses = _get_addresses_for_peers(site["peers"],
                                            [ipaddress.ip_network(n) for n in site["ip_networks"]])

  prepare_directory(wg_config_path)

  created_files = []
  for p in site["peers"]:
    created_files.append(
        write_file(
            os.path.join(wg_config_path, f"wg_{p}.conf"),
            _get_peer_config(p, site["peers"], peer_addresses)))
  return created_files


def delete_config(site_name: str, wg_config_path: str):
  """ Delete the config files for a site """

  delete_directory(os.path.join(wg_config_path))


def _get_peer_config(interface_peer_name: str, peers: Peers, peer_addresses: dict):
  """ Write config file for a peer """
  s = _get_interface_section(interface_peer_name, peers[interface_peer_name], peer_addresses)

  for p in peers:
    if p != interface_peer_name and ((p in peers[interface_peer_name]["ingoing_connected_peers"]) or (p in peers[interface_peer_name]["outgoing_connected_peers"])):
      s += _get_peer_section(p, peers[p], interface_peer_name, peers[interface_peer_name], peer_addresses)

  return s


def _get_interface_section(name: str, peer: PeerItems, peer_addresses: dict) -> str:
  """ Get the interface section of a config file for a peer"""

  s = f"# {name}\n"
  s += "[Interface]\n"
  s += _get_address_line(name, peer_addresses)
  if peer["ingoing_connected_peers"]:
    s += f"ListenPort = {peer['port']}\n"
    #TODO: firewall rules
  if peer["redirect_all_traffic"]:
    s += f"DNS = 1.1.1.1, 8.8.8.8\n"
  s += f"PrivateKey = " + peer["keys"]["privkey"] + "\n"
  s += "\n"
  return s


def _get_peer_section(name: str, peer: PeerItems, interface_peer_name: str, interface_peer: PeerItems, peer_addresses: dict):
  """ Get the peer section of a config file """

  s = f"# {name}\n"
  s += f"[Peer]\n"
  if peer["endpoint"] and name in interface_peer["outgoing_connected_peers"]:
    s += f"Endpoint = {peer['endpoint']}:{peer['port']}\n"
    if peer["persistent_keep_alive"]:
      s += "PersistentKeepAlive = 25\n"
  s += f"PublicKey = " + peer["keys"]["pubkey"] + "\n"

  # Always the psk of the outgoing_connected_peers is used
  # If a peer is ingoing and outgoing the psk of the alphebetically first peer is used
  if name in interface_peer["ingoing_connected_peers"] and name in interface_peer["outgoing_connected_peers"]:
    l = [name, interface_peer_name]
    l.sort()
    if name == l[0]:
      s += f"PresharedKey = " + peer["keys"]["psk"] + "\n"
    else:
      s += f"PresharedKey = " + interface_peer["keys"]["psk"] + "\n"
  elif name in interface_peer["ingoing_connected_peers"]:
    s += f"PresharedKey = " + peer["keys"]["psk"] + "\n"
  else:
    # Peer has to be outgoing_connected_peer
    s += f"PresharedKey = " + interface_peer["keys"]["psk"] + "\n"
  s += _get_allowed_ips_line(
      peer_name=name,
      peer=peer,
      interface_peer=interface_peer,
      peer_addresses=peer_addresses,
      redirect_all_traffic=interface_peer["redirect_all_traffic"])
  s += "\n"
  return s


def _get_addresses_for_peers(peers: tuple,
                             ip_networks: list):
  """ Create ip addresses for each peer """

  peer_addresses = {}
  address_iterators = []
  for n in ip_networks:
    address_iterators.append([n, n.hosts()])
  for p in peers:
    peer_addresses.update({p: {}})
    for i in address_iterators:
      peer_addresses[p].update({i[0]: next(i[1])})
  return peer_addresses


def _get_address_line(peer_name: str, peer_addresses: dict):
  """ Create the Address line """

  address_line = "Address = "
  for network in peer_addresses[peer_name].keys():
    address_line += str(peer_addresses[peer_name][network]) + "/" + str(
        network.prefixlen) + ", "

  return address_line[:-2] + "\n"


def _get_allowed_ips_line(peer_name: str, peer: PeerItems, interface_peer: PeerItems, peer_addresses: dict,
                          redirect_all_traffic: bool):
  """ Create the AllowedIPs line """

  allowed_ips_line = "AllowedIPs = "
  for network in peer_addresses[peer_name].keys():
    if redirect_all_traffic and peer_name in interface_peer["main_peer"]:
      if network.version == 4:
        allowed_ips_line += "0.0.0.0/0, "
      else:
        allowed_ips_line += "::/0, "
    else:
      allowed_ips_line += str(peer_addresses[peer_name][network]) + "/" + str(
          network.max_prefixlen) + ", "
    for s in peer["additional_allowed_ips"]:
      allowed_ips_line += s + ", "


  return allowed_ips_line[:-2] + "\n"



import ipaddress

from .console import print_message
from .console import print_error

from .shared import create_wireguard_config
from .shared import get_new_peer_properties
from .shared import yes_no_menu

from ..library import edit_connection_table
from ..library import edit_string
from ..library import ConnectionTable
from ..library import Peer
from ..library import Site
from ..library import SiteDoesExistError
from ..library import SiteDoesNotExistError
from ..library import WireUI


def list_sites(w: WireUI) -> int:
  sites = w.get_sites()
  if sites:
    print_message(0, "The following sites do exist:")
    for s in sites:
      print_message(0, s)
  else:
    print_message(0, "There are currently no sites.")
  return len(sites)


def add_site(w: WireUI) -> str:
  site_name = get_site_name(w, should_exist=False)

  ip = []
  if yes_no_menu("Do you want to use IPv4?"):
    ip.append(_get_ip_network(4))
  if yes_no_menu("Do you want to use IPv6?"):
    ip.append(_get_ip_network(6))

  peer_names = _get_peer_names()

  # Editing of the connection table
  ct = ConnectionTable(peer_names)
  input("Please edit the connection table. Press ENTER to continue...")
  ct = edit_connection_table(ct)

  # Create peer list
  peers = []
  for p in peer_names:
    peers.append(get_new_peer_properties(w, site_name, p, ct))

  try:
    w.add_site(
        Site(site_name, ip, peers))
  except SiteDoesExistError as e:
    print_error(0, "Site does already exist. Doing nothing...")
    print_error(0, e)

  create_wireguard_config(w, site_name)

  return site_name


def delete_site(w: WireUI):
  site_name = get_site_name(w, should_exist=True)
  if yes_no_menu("Do you really want to delete the site \"" + site_name +
                 "\"?"):
    w.delete_wireguard_config(site_name)
    try:
      w.delete_site(site_name)
    except SiteDoesNotExistError as e:
      print_error(0, "Site does not exist. Doing nothing...")
      print_error(0, e)


def get_site_name(w: WireUI, should_exist: bool) -> str:
  while True:
    name = input("Please enter the name of the site: ")
    exist = w.site_exists(name)
    if (should_exist and exist) or (not should_exist and not exist):
      return name
    elif should_exist and not exist:
      print_error(0, f"Error: {name} does not exist.")
    elif not should_exist and exist:
      print_error(0, f"Error: {name} does already exist.")


# TODO: is correct messages
def _get_ip_network(ip_version: int = 4) -> str:
  while True:
    if ip_version == 4:
      s = "Please enter the IPv4 network the site should use.\n(For example \"10.0.0.0/24\"): "
    else:
      s = "Please enter the IPv6 network the site should use.\n(For example \"fd01::/64\"): "

    ip_network = input(s)

    try:
      ip_network = ipaddress.ip_network(ip_network)
    except ValueError as e:
      print_error(0, "Error: Input is not a valid IP network.")
      print_error(2, e)
    else:
      if ip_version != ip_network.version:
        print_error(
            0, "Error: You entered an IPv" + ip_network.version +
            " network, but an IPv" + ip_version + " network is expected.")
        continue

      return str(ip_network.with_prefixlen)


def _get_peer_names() -> tuple:
  # Get peer names
  peer_names = input("Please enter the name of the peers (use ' ' as separation): ")
  peer_names = _convert_str_to_list(peer_names)

  # Check names
  peer_names = _convert_list_to_str(peer_names)

  correct = False
  while not correct:
    correct = yes_no_menu("Is everything correct?")
    if not correct:
      peer_names = edit_string(peer_names)
      peer_names = _convert_str_to_list(peer_names)
      peer_names = _convert_list_to_str(peer_names)

  peer_names = _convert_str_to_list(peer_names)
  return peer_names

def _convert_str_to_list(peer_names: str) -> list:
  """ Convert a str to a list.

  Doubled entires are removed and the list is sorted. """

  peer_names = list(set(peer_names.split()))
  peer_names.sort()
  return peer_names

def _convert_list_to_str(peer_names: list) -> str:
  """ Convert a list to a str.

  Elements are separated with ' '"""

  print_message(0, "The following peers has been detected:")
  s = ""
  for p in peer_names:
    print_message(0, str(p))
    s += f"{p} "
  s = s[:-1]

  return s



from atexit import register

from .console import set_verbosity

from .menus import start_message
from .menus import entrypoint_menu

from .shared import print_error

from ..library import DataIntegrityError
from ..library import WireUI
from ..library import SettingDoesNotExistError

_SETTINGS_PATH = "./settings.json"


def run_ui():
  s = ""
  try:
    w = WireUI("./settings.json")
  except DataIntegrityError as e:
    s += f"{e}\n"
  if s:
    print_error(0, s[:-1])
    return

  register(w.write_settings_to_file)
  register(w.write_sites_to_file)
  set_verbosity(w.get_setting("verbosity"))
  start_message()
  entrypoint_menu(w)



from .config import delete_config
from .config import write_config

from .io_ import edit_connection_table
from .io_ import edit_dict
from .io_ import edit_string
from .io_ import read_file
from .io_ import write_file

from .keys import get_keys

from .typedefs import ConnectionTable
from .typedefs import DataIntegrityError
from .typedefs import JSONDecodeError
from .typedefs import JsonDict
from .typedefs import KeyDoesExistError
from .typedefs import KeyDoesNotExistError
from .typedefs import Keys
from .typedefs import PeerDoesExistError
from .typedefs import PeerDoesNotExistError
from .typedefs import PeerItems
from .typedefs import Peers
from .typedefs import SettingDoesExistError
from .typedefs import SettingDoesNotExistError
from .typedefs import Settings
from .typedefs import SiteDoesExistError
from .typedefs import SiteDoesNotExistError
from .typedefs import SiteItems
from .typedefs import Sites
from .typedefs import WireguardNotFoundError

from .wireui import Site
from .wireui import Peer
from .wireui import WireUI

__all__ = [
    "delete_config",
    "edit_connection_table",
    "edit_dict",
    "edit_string",
    "get_keys",
    "read_file",
    "write_config",
    "write_file",
    "ConnectionTable",
    "DataIntegrityError",
    "JSONDecodeError",
    "JsonDict",
    "Keys",
    "KeyDoesExistError",
    "KeyDoesNotExistError",
    "Peer",
    "PeerDoesExistError",
    "PeerDoesNotExistError",
    "PeerItems",
    "Peers",
    "Settings",
    "SettingDoesExistError",
    "SettingDoesNotExistError",
    "Site",
    "SiteDoesExistError",
    "SiteDoesNotExistError",
    "SiteItems",
    "Sites",
    "WireguardNotFoundError",
    "WireUI",
]



# integrity.py
# Check data integrity
# Author: Tim Schlottmann

import ipaddress

from .typedefs import DataIntegrityError
from .typedefs import PeerItems
from .typedefs import Peers
from .typedefs import Settings
from .typedefs import SiteItems
from .typedefs import Sites

version_dict = {
  "0.1.0": 1
}

settings_latest_version = "0.1.0"
site_latest_version = "0.1.0"


# Data check recipe
#
# A check is done after the following recipe:
# 1. Check if a key is present
# 2. Check if the value has the correct datatype
# 3. If the value is of type list or dict repeat step 2
# 4. Make additional checks where applicable

def check_site_integrity(sites: Sites):
  """ Check data integrity of the sites """
  for s in sites:
    # Update routines for old file versions
    _check_key(sites[s], "version", "site", s, str, "str")
    if sites[s]["version"] not in version_dict:
      raise DataIntegrityError(f"Site {s} is version {sites[s]['version']}, which is currently not supported. Latest supported version is {site_latest_version}")
    if version_dict[sites[s]["version"]] < version_dict[site_latest_version]:
      # Put update routines here
      pass
    elif version_dict[sites[s]["version"]] > version_dict[site_latest_version]:
      raise DataIntegrityError(f"Site {s} is version {sites[s]['version']}, which is currently not supported. Latest supported version is {site_latest_version}")

    # Data integrity check

    #Flags
    ipv4_network = False
    ipv6_network = False
    version = sites[s]["version"]

    # Check ip_networks
    _check_key(sites[s], "ip_networks", "site", s, list, "list")
    for n in sites[s]["ip_networks"]:
      _check_datatype(n, "key", "ip_networks", str, "str" )
      try:
        v = ipaddress.ip_network(n).version
      except Exception as e:
        raise DataIntegrityError(f"{n} in key \"ip_networks\" in site {s} is not a valid IP network\n{e}")
      if v == 4:
        ipv4_network = True
      elif v == 6:
        ipv6_network = True

    # Check peers
    check_peer_integrity(Peers(sites[s]["peers"]), s, ipv4_network, ipv6_network, version)


def check_peer_integrity(peers: Peers, site_name: str, ipv4_network: bool, ipv6_network: bool, version: str):
    for p in peers:
      # Update routines for old file versions
      if version_dict[version] < version_dict[site_latest_version]:
        # Put update routines here
        pass

      # Data integrity check

      # Flags

      # Check keys
      _check_key(peers[p], "keys", "peer", f"{p} (site \"{site_name}\")", dict, "dict")
      for k in peers[p]["keys"]:
       _check_key(peers[p]["keys"], k, "keys", k, str, "str")

      # Check additional allowed ips
      _check_key(peers[p], "additional_allowed_ips", "peer", f"{p} (site \"{site_name}\")", list, "list")
      for n in peers[p]["additional_allowed_ips"]:
        try:
          v = ipaddress.ip_network(n).version
        except Exception as e:
          raise DataIntegrityError(f"{n} in key \"additional_allowed_ips\" in peer {p} (site \"{site_name}\") is not a valid IP network\n{e}")
        if v == 4 and not ipv4_network:
          raise DataIntegrityError(f"{n} in key \"additional_allowed_ips\" in peer {p} (site \"{site_name}\") is IPv4, but IPv4 is not activated in the site")
        elif v == 6 and not ipv6_network:
          raise DataIntegrityError(f"{n} in key \"additional_allowed_ips\" in peer {p} (site \"{site_name}\") is IPv6, but IPv6 is not activated in the site")

      # Check ingoing and outgoing_connected_peers and main_peer
      # If peer p2 is outgoing_connected_peer of peer p1, p1 has to be ingoing_connected of peer p2.
      _check_key(peers[p], "outgoing_connected_peers", "peer",f"{p} (site \"{site_name}\")", list, "list")
      _check_key(peers[p], "main_peer", "peer", f"{p} (site \"{site_name}\")", str, "str")
      _check_key(peers[p], "ingoing_connected_peers", "peer",f"{p} (site \"{site_name}\")", list, "list")

      for outgoing_peer in peers[p]["outgoing_connected_peers"]:
        if outgoing_peer not in peers:
          raise DataIntegrityError(f"{outgoing_peer} is set as outgoing peer in peer {p} (site \"{site_name}\"), but is not present in peers-list.")
        if p not in peers[outgoing_peer]["ingoing_connected_peers"]:
          raise DataIntegrityError(f"{outgoing_peer} is set as outgoing peer in peer {p} (site \"{site_name}\"), but is not present as ingoing_connected_peer in peer {outgoing_peer}")
      for ingoing_peer in peers[p]["ingoing_connected_peers"]:
        if ingoing_peer not in peers:
          raise DataIntegrityError(f"{ingoing_peer} is set as ingoing peer in peer {p} (site \"{site_name}\"), but is not present in peers-list.")
        if p not in peers[ingoing_peer]["outgoing_connected_peers"]:
          raise DataIntegrityError(f"{ingoing_peer} is set as ingoing peer in peer {p} (site \"{site_name}\"), but is not present as ingoing_connected_peer in peer {ingoing_peer}")
      if peers[p]["outgoing_connected_peers"] and (peers[p]["main_peer"] not in peers[p]["outgoing_connected_peers"]):
          raise DataIntegrityError(f"{peers[p]['main_peer']} in peer {p} (site \"{site_name}\") is mentioned as main_peer, but is not present as outgoing_connected_peer")

      # Check endpoint
      _check_key(peers[p], "endpoint", "peer", f"{p} (site \"{site_name}\")", str, "str")

      # Check port
      _check_key(peers[p], "port", "peer", f"{p} (site \"{site_name}\")", int, "int")

      # Check redirect_all_traffic
      _check_key(peers[p], "redirect_all_traffic", "peer", f"{p} (site \"{site_name}\")", bool, "bool")


def check_settings_integrity(settings: Settings):
    # Update routines for old file versions
    _check_key(settings, "file_version", "key", "settings", str, "str")
    if settings["file_version"] not in version_dict:
      raise DataIntegrityError(f"The settings are version {settings['file_version']}, which is currently not supported. Latest supported version is {settings_latest_version}")
    if version_dict[settings["file_version"]] < version_dict[settings_latest_version]:
      # Put update routines here
      pass
    elif version_dict[settings["file_version"]] > version_dict[settings_latest_version]:
      raise DataIntegrityError(f"The settings are version {settings['file_version']}, which is currently not supported. Latest supported version is {settings_latest_version}")

    # Data integrity check

    # Flags

    # Check keys
    _check_key(settings, "verbosity", "key", "settings", int, "int")
    _check_key(settings, "sites_file_path", "key", "settings", str, "str")
    _check_key(settings, "wg_config_path", "key", "settings", str, "str")
    _check_key(settings, "editor", "key", "settings", str, "str")


def _check_key(d: dict, key: str, what_is_checked: str, dict_name: str, datatype: type, datatype_name: str):
  _check_key_presence(d, key, what_is_checked, dict_name, datatype, datatype_name)
  _check_datatype(d[key], what_is_checked, dict_name, datatype, datatype_name)


def _check_key_presence(d: dict, key: str, what_is_checked: str, dict_name: str, datatype: type, datatype_name: str):
  try:
    d[key]
  except KeyError:
    raise DataIntegrityError(f"Key \"{key}\" not present in {what_is_checked} \"{dict_name}\"")


def _check_datatype(o, what_is_checked: str, dict_name: str, datatype: type, datatype_name: str):
  if not isinstance(o, datatype):
    raise DataIntegrityError(f"\"{o}\" in {what_is_checked} \"{dict_name}\" should be {datatype_name} not {type(o)}")



# database.py
# Basic database
# Author: Tim Schlottmann

from collections import UserDict
from json import dumps
from json import loads
from json import JSONDecodeError
from typing import Dict
from typing import Union

# DataDict = Dict[str, dict]

# StringDict = Dict[str, str]

# JsonString = str


class JsonDict(UserDict):
  def __init__(self,
               initialdata: Union[dict, str] = {},
               defaults: Union[dict, str] = {}):
    d = self._getdict(defaults)
    d.update(self._getdict(initialdata))
    super().__init__(d)

  @classmethod
  def _getdict(cls, data: Union[dict, str]) -> dict:
    if isinstance(data, dict):
      return data
    elif isinstance(data, str):
      if not data:
        data = "{}"
      try:
        return loads(data)
      except JSONDecodeError as e:
        raise e
    else:
      raise TypeError(
          f"Excpected type of data is dict or str. Got {type(data)}")

  def __repr__(self):
    return f"{type(self).__name__}({self.data})"

  def __str__(self):
    return dumps(self.data, indent=2)



import ipaddress

from .console import print_error
from .console import print_message

from .shared import create_wireguard_config
from .shared import get_additional_allowed_ips
from .shared import get_endpoint
from .shared import get_new_peer_properties
from .shared import get_port
from .shared import get_persistent_keep_alive
from .shared import get_redirect_all_traffic
from .shared import yes_no_menu

from ..library import edit_connection_table
from ..library import ConnectionTable
from ..library import Peer
from ..library import PeerDoesExistError
from ..library import PeerDoesNotExistError
from ..library import WireUI


def list_peers(w: WireUI, site_name: str) -> int:
  peers = w.get_peer_names(site_name)
  if peers:
    print_message(0, "The following peers already exist:")
    for p in peers:
      print_message(0, p)
  else:
    print_message(0, "There are currently no peers.")
  return len(peers)


def add_peer(w: WireUI, site_name: str):
  peer_name = _get_peer_name(w, site_name, should_exist=False)
  try:
    w.add_peer(site_name, get_new_peer_properties(w, site_name, peer_name, ConnectionTable([peer_name])))
  except PeerDoesExistError:
    print_error(0, "Error: Peer does already exist. Do nothing...")
  else:
    edit_peer_connections(w, site_name)
    create_wireguard_config(w, site_name)


# TODO: add the new features --> transition this function to edit_peer
def rekey_peer(w: WireUI, site_name: str):
  peer_name = _get_peer_name(w, site_name, should_exist=True)
  try:
    w.rekey_peer(site_name, peer_name)
  except PeerDoesNotExistError:
    print_error(0, "Error: Peer does not exist. Do nothing...")
  else:
    create_wireguard_config(w, site_name)


def delete_peer(w: WireUI, site_name: str):
  peer_name = _get_peer_name(w, site_name, should_exist=True)
  if yes_no_menu(f"Do you really want to delete peer {peer_name}?"):
    try:
      w.delete_peer(site_name, peer_name)
    except PeerDoesNotExistError:
      print_error(0, "Error: Peer does not exist. Do nothing...")

  edit_peer_connections(w, site_name)
  create_wireguard_config(w, site_name)


def edit_peer_connections(w: WireUI, site_name: str):
  """ Edit the peer connection matrix """

  # Create an empty connection table
  peer_names = w.get_peer_names(site_name)
  ct = ConnectionTable(peer_names)

  # Populate table with actual data
  for i in range(len(peer_names)):
    peer = w.get_peer(site_name, ct.row_names[i])
    for j in range(len(peer_names)):
      if ct.column_names[j] in peer.outgoing_connected_peers:
        ct.setitem(i, j, 1)
    ct.setitem(i, len(peer_names), peer.main_peer)

  # Edit table
  input("Please edit the connection table. Press ENTER to continue...")
  ct = edit_connection_table(ct)

  # Update peers with changed connection table
  for p in peer_names:
    peer_old = w.get_peer(site_name, p)

    # If a peer now has ingoing connections, ask for endpoint and port
    if not peer_old.ingoing_connected_peers and ct.get_ingoing_connected_peers(p):
      print_message(0, f"Collecting properties of peer {p} (ingoing)...")
      if peer_old.endpoint == "":
        endpoint = get_endpoint()
      else:
        endpoint = peer_old.endpoint
      if peer_old.port == 0:
        port = get_port()
      else:
        port = peer_old.port
      if peer_old.persistent_keep_alive == -1:
        persistent_keep_alive = get_persistent_keep_alive()
      else:
        persistent_keep_alive = peer_old.persistent_keep_alive
      if peer_old.additional_allowed_ips == []:
        allow_ipv4 = False
        allow_ipv6 = False
        for n in w.get_networks(site_name):
          v = ipaddress.ip_network(n).version
          if v == 4:
            allow_ipv4 = True
          elif v == 6:
            allow_ipv6 = True
        additional_allowed_ips = get_additional_allowed_ips(allow_ipv4, allow_ipv6)
      else:
        additional_allowed_ips = []
    else:
      endpoint = peer_old.endpoint
      port = peer_old.port
      persistent_keep_alive = peer_old.persistent_keep_alive
      additional_allowed_ips = peer_old.additional_allowed_ips

    # If a peer now has outgoing connections, ask for redirect_all_traffic
    if not peer_old.outgoing_connected_peers and ct.get_outgoing_connected_peers(p):
      print_message(0, f"Collecting properties of peer {p} (outgoing)...")
      if peer_old.redirect_all_traffic == None:
        redirect_all_traffic = get_redirect_all_traffic()
      else:
        redirect_all_traffic = peer_old.redirect_all_traffic
    else:
      redirect_all_traffic = peer_old.redirect_all_traffic

    w.set_peer(site_name, Peer(
        peer_old.name,
        additional_allowed_ips,
        ct.get_outgoing_connected_peers(p),
        ct.get_main_peer(p),
        ct.get_ingoing_connected_peers(p),
        endpoint,
        port,
        persistent_keep_alive,
        redirect_all_traffic,
    ))

  create_wireguard_config(w, site_name)


def _get_peer_name(w: WireUI, site_name: str, should_exist: bool) -> str:
  while True:
    peer_name = input("Please enter the name of the peer: ")
    exist = w.peer_exists(site_name, peer_name)
    if (should_exist and exist) or (not should_exist and not exist):
      return peer_name
    elif should_exist and not exist:
      print_error(0, f"Error: {peer_name} does not exist.")
    elif not should_exist and exist:
      print_error(0, f"Error: {peer_name} does already exist.")



# tables.py
# Table for wireguard
# Author: Tim Schlottmann

from .exceptions import PeerDoesNotExistError

class Table():
  """ n x m table """

  def __init__(self, n: int , m: int, row_names: list = [], column_names: list = []):
    if len(row_names) != n and row_names != []:
      raise ValueError("Dimension mismatch: len(row_names) != n")
    if len(column_names) != m and column_names != []:
      raise ValueError("Dimension mismatch: len(column_names) != m")

    self.n = n
    self.m = m

    # Create n x m matrix
    self.content = [None] * self.n
    for i in range(self.n):
      self.content[i] = [None] * self.m

    self.row_names = row_names
    self.column_names = column_names

    # Get parameters from names
    if self.row_names != []:
      self.row_names_lengths = [0] * n
      self.row_names_max_length = 0
      for i in range(n):
        self.row_names_lengths[i] = len(self.row_names[i])
        if self.row_names_lengths[i] > self.row_names_max_length:
          self.row_names_max_length = self.row_names_lengths[i]

    if self.column_names != []:
      self.column_names_lengths = [0] * m
      self.column_names_max_length = 0
      for i in range(m):
        self.column_names_lengths[i] = len(self.column_names[i])
        if self.column_names_lengths[i] > self.column_names_max_length:
          self.column_names_max_length = self.column_names_lengths[i]

  def __repr__(self):
    return f"{type(self).__name__}({self.n}, {self.m}, {self.row_names}, {self.column_names})"

  def __str__(self):
    s = ""

    # Print column headings (only if there are any)
    if len(self.column_names) > 0:
      for i in range(self.row_names_max_length):
        s += " "
      for i in range(len(self.column_names)):
        s += f" {self.column_names[i]}"
      s += "\n"

    # Print rows (only if there are any)
    for i in range(self.n):
      if len(self.row_names) > 0:
        for j in range(self.row_names_max_length - self.row_names_lengths[i]):
          s += " "
        s += self.row_names[i]
      for j in range(self.m):
        s += f" {self.content[i][j]}"
        if len(self.column_names) > 0 and len(str(self.content[i][j])) < self.column_names_lengths[j]:
          for _ in range(self.column_names_lengths[j] - len(str(self.content[i][j]))):
            s += " "
      s += "\n"

    #Remove last new line
    s = s[:-1]

    return s

  def getitem(self, i: int, j: int) -> any:
    """ Get the value of the item in row i and column j """
    return self.content[i][j]

  def setitem(self, i: int, j: int, v: any):
    """ Set the item in row i and column j to value v """
    self.content[i][j] = v

  def setrow(self, i: int, r: list):
    if len(r) == self.m:
      self.content[i] = r
    else:
      raise ValueError(f"Dimension mismatch: len(r) ({len(r)}) != self.m ({self.m})")

class ConnectionTable(Table):
  """ ConnectionTable for peers """
  def __init__(self, peer_names: list):
    super().__init__(len(peer_names), len(peer_names) + 1, peer_names, peer_names + ["main_peer"])
    for i in range(self.n):
      for j in range(self.n):
        self.setitem(i, j, 0)
      self.setitem(i, self.m - 1, "None")

  def __repr__(self):
    return f"{type(self).__name__}({self.row_names})"


  def setitem(self, i: int, j: int, v: any):
    """ Set the item in row i and column j to value v

        Please execute check_integrity afterwards to make sure changed data is still valid """

    if i == j and v == 1:
      raise ValueError("A peer cannot be connected to itself. Please make sure that all diagonal elements are '0'!")
    if j == self.m - 1 and not isinstance(v, str):
      raise ValueError(f"v ({v}) has to be an str not {type(v)}")

    super().setitem(i, j, v)


  def update(self, s: str):
    """ Updates the table with a str representation of a ConnectionTable object """

    # Split lines and remove first line
    s = s.splitlines()
    s.pop(0)

    for i in range(self.n):
      # Separate connection elements
      s[i] = s[i][self.row_names_max_length + 1:]
      s[i] = s[i].split()

      if len(s[i]) != self.m:
        raise ValueError(f"Dimension mismatch: len(s[i]) ({len(s[i])}) != self.m ({self.m})")

      # Update table
      for j in range(self.m):
        if j < self.m - 1:
          self.setitem(i, j, int(s[i][j]))
        else:
          self.setitem(i, j, s[i][j])

    self._check_integrity()


  def get_outgoing_connected_peers(self, name: str) -> list:
    """ Get a list of all peers that peer 'name' has an outgoing connection to """
    # Get row index for peer
    row = -1
    for i in range(self.n):
      if self.column_names[i] == name:
        row = i
        break
    if row == -1:
      raise PeerDoesNotExistError(name)

    # List all peers with an outgoing connection to that peer
    l = []
    for i in range(self.n):
      if self.getitem(row, i):
        l.append(self.column_names[i])

    return l


  def get_main_peer(self, name: str) -> str:
    """ Get the main peer for outgoing connections for a peer """
    # Get row index for peer
    row = -1
    for i in range(self.n):
      if self.column_names[i] == name:
        row = i
        break
    if row == -1:
      raise PeerDoesNotExistError(name)

    return (self.getitem(row, self.m - 1))


  def get_ingoing_connected_peers(self, name: str) -> list:
    """ Get a list of all peers that peer 'name' has an ingoing connection from """
    # Get column index for peer
    column = -1
    for i in range(self.m):
      if name == self.column_names[i]:
        column = i
        break

    if column == -1:
      raise PeerDoesNotExistError(name)

    # List all peers with an ingoing connection from that peer
    l = []
    for i in range(self.n):
      if self.getitem(i, column):
        l.append(self.row_names[i])

    return l

  def _check_integrity(self):
    """ Check if the table has invalid entries """

    msg = ""

    #Check if all diagonal elements are zero
    for i in range(self.n):
      if self.getitem(i, i) == 1:
        self.setitem(i, i, 0)
        msg += f"Peer {self.row_names[i]}: A peer cannot be connected to itself. Please make sure that all diagonal elements are '0'!\n"

    #Check if there is an main peer if there is an outgoing connection
    for i in range(self.n):
      if len(self.get_outgoing_connected_peers(self.row_names[i])) > 0:
        if not self.getitem(i, self.m - 1):
          msg += f"Peer {self.row_names[i]}: If there is an outgoing connection there has to be a main_peer.\n"
          self.setitem(i, self.m - 1, self.get_outgoing_connected_peers(self.row_names[i])[0])
        elif self.getitem(i, self.m - 1) not in self.get_outgoing_connected_peers(self.row_names[i]):
          msg += f"Peer {self.row_names[i]}: The main_peer is not an outgoing_connected_peer.\n"

    if msg:
      raise ValueError(msg[:-1])



#!/usr/bin/python3

# wireui
# A tool for creating and managing wireguard configs
# Version 0.1.0a
# (C) 2020 Tim Schlottmann

# import ui.menus
# import inputOutput.console as console
# from database.settings import settings

from .ui import run_ui


def main():
  run_ui()


if __name__ == "__main__":
  main()



# wire_interface.py
# Get keys from wg Write wireguard config files
# Author: Tim Schlottmann

from os import path
from typing import NamedTuple
from typing import Optional

# from .io_ import JSONDecodeError

from .config import delete_config
from .config import write_config

from .integrity import check_settings_integrity
from .integrity import check_site_integrity
from .integrity import settings_latest_version
from .integrity import site_latest_version

from .io_ import read_file
from .io_ import write_file

from .keys import get_keys

from .typedefs import JSONDecodeError
from .typedefs import PeerItems
from .typedefs import PeerDoesExistError
from .typedefs import PeerDoesNotExistError
from .typedefs import Peers
from .typedefs import Settings
from .typedefs import SettingDoesNotExistError
from .typedefs import SiteItems
from .typedefs import SiteDoesExistError
from .typedefs import SiteDoesNotExistError
from .typedefs import Sites


class Site(NamedTuple):
  name: str
  ip_networks: str
  peers: list


class Peer(NamedTuple):
  name: str
  additional_allowed_ips: list
  outgoing_connected_peers: list
  main_peer: str
  ingoing_connected_peers: list
  endpoint: str
  port: int
  persistent_keep_alive: int
  redirect_all_traffic: bool


class WireUI():
  """ Class for managing wireguard config files """

  def __init__(self, settings_path: Optional[str] = None):
    default_settings = {
        "file_version": settings_latest_version,
        "verbosity": 0,
        "sites_file_path": "./sites.json",
        "wg_config_path": "./wg",
        "editor": "editor",
    }
    if settings_path:
      self.settings_path = settings_path
      try:
        self._settings = Settings(read_file(settings_path), default_settings)
      except JSONDecodeError as e:
        raise e
    else:
      self._settings = Settings(defaults=default_settings)

    self._sites = Sites(read_file(self._settings.get("sites_file_path")))

    check_settings_integrity(self._settings)
    check_site_integrity(self._sites)

  def get_sites(self) -> list:
    """ Get all existing sites """

    return list(self._sites)

  def add_site(self, site: Site):
    """ Add a new site and creates its config files """

    if site.name in self._sites:
      raise SiteDoesExistError(site.name)

    peers = Peers()
    for p in site.peers:
      try:
        peers[p.name] = PeerItems({
          "keys": get_keys(),
          "additional_allowed_ips": p.additional_allowed_ips,
          "outgoing_connected_peers": p.outgoing_connected_peers,
          "main_peer": p.main_peer,
          "ingoing_connected_peers": p.ingoing_connected_peers,
          "endpoint": p.endpoint,
          "port": p.port,
          "persistent_keep_alive": p.persistent_keep_alive,
          "redirect_all_traffic": p.redirect_all_traffic,
        })
      except PeerDoesExistError as e:
        raise e

    self._sites[site.name] = SiteItems({
        "config_version": site_latest_version,
        "ip_networks": site.ip_networks,
        "peers": peers
    })

  def delete_site(self, name: str):
    """ Delete a site """

    if name not in self._sites:
      raise SiteDoesNotExistError(name)

    del self._sites[name]

  def site_exists(self, name: str) -> bool:
    """ Check if a site does exist """

    return name in self._sites

  def get_number_of_sites(self) -> int:
    """ Get the number of existing sites """

    return len(self._sites)

  def get_peer_names(self, site_name: str) -> list:
    """ Get name of all existing peers from a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    return list(self._sites[site_name]["peers"])

  def add_peer(self, site_name: str, peer: Peer):
    """ Add a peer to a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    if peer.name in self._sites[site_name]["peers"]:
      raise PeerDoesExistError(peer.name)

    self._sites[site_name]["peers"][peer.name] = PeerItems({
        "keys": get_keys(),
        "additional_allowed_ips": peer.additional_allowed_ips,
        "outgoing_connected_peers": peer.outgoing_connected_peers,
        "main_peer": peer.main_peer,
        "ingoing_connected_peers": peer.ingoing_connected_peers,
        "endpoint": peer.endpoint,
        "port": peer.port,
        "persistent_keep_alive": peer.persistent_keep_alive,
        "redirect_all_traffic": peer.redirect_all_traffic,
    })

  def get_peer(self, site_name: str, peer_name: str) -> Peer:
    """ Get a peer from a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    if peer_name not in self._sites[site_name]["peers"]:
      raise PeerDoesNotExistError(peer_name)

    return Peer(
      peer_name,
      self._sites[site_name]["peers"][peer_name]["additional_allowed_ips"],
      self._sites[site_name]["peers"][peer_name]["outgoing_connected_peers"],
      self._sites[site_name]["peers"][peer_name]["main_peer"],
      self._sites[site_name]["peers"][peer_name]["ingoing_connected_peers"],
      self._sites[site_name]["peers"][peer_name]["endpoint"],
      self._sites[site_name]["peers"][peer_name]["port"],
      self._sites[site_name]["peers"][peer_name]["persistent_keep_alive"],
      self._sites[site_name]["peers"][peer_name]["redirect_all_traffic"],
    )

  def set_peer(self, site_name: str, peer: Peer):
    """ Set a peer in a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    if peer.name not in self._sites[site_name]["peers"]:
      raise PeerDoesNotExistError(peer.name)

    self._sites[site_name]["peers"][peer.name] = PeerItems({
        "keys": self._sites[site_name]["peers"][peer.name]["keys"],
        "additional_allowed_ips": peer.additional_allowed_ips,
        "outgoing_connected_peers": peer.outgoing_connected_peers,
        "main_peer": peer.main_peer,
        "ingoing_connected_peers": peer.ingoing_connected_peers,
        "endpoint": peer.endpoint,
        "port": peer.port,
        "persistent_keep_alive": peer.persistent_keep_alive,
        "redirect_all_traffic": peer.redirect_all_traffic,
    })

  def delete_peer(self, site_name: str, peer_name: str):
    """ Delete a peer from a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    if peer_name not in self._sites[site_name]["peers"]:
      raise PeerDoesNotExistError(peer_name)

    del self._sites[site_name]["peers"][peer_name]

  def rekey_peer(self, site_name: str, peer_name: str):
    """ Create new keys for a peer from a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    if peer_name not in self._sites[site_name]["peers"]:
      raise PeerDoesNotExistError(peer_name)

    self._sites[site_name]["peers"][peer_name] = PeerItems({
        "keys": get_keys()
    })

  def peer_exists(self, site_name: str, peer_name: str) -> bool:
    """ Check if a peer exists in a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    return peer_name in self._sites[site_name]["peers"]

  def get_number_of_peers(self, site_name: str) -> int:
    """ Get the number of peers in a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)
    return len(self._sites[site_name]["peers"])


  def get_networks(self, site_name: str) -> list:
    return self._sites[site_name]["ip_networks"]

  def create_wireguard_config(self, site_name: str) -> list:
    """ Write the wireguard config files """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    return write_config(self._sites[site_name],
                        path.join(self._settings["wg_config_path"], site_name))

  def delete_wireguard_config(self, site_name: str):
    """ Check if a peer exists in a site """

    if site_name not in self._sites:
      raise SiteDoesNotExistError(site_name)

    delete_config(site_name,
                  path.join(self._settings.get("wg_config_path"), site_name))

  def get_setting_names(self, setting: str) -> list:
    """ Get names of all existing settings """

    return list(self._settings)

  def get_setting(self, setting: str):
    """ Get a setting """

    if setting not in self._settings:
      raise SettingDoesNotExistError

    return self._settings[setting]

  def write_settings_to_file(self):
    write_file(self.settings_path, str(self._settings))

  def write_sites_to_file(self):
    write_file(self._settings["sites_file_path"], str(self._sites))



import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
    name="wireui-TheTimmoth",
    version="0.1.0a3",
    author="Tim Schlottmann",
    author_email="tschlottmann@gmail.com",
    description="A tool for creating and managing wireguard configs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheTimmoth/wireui",
    packages=setuptools.find_packages(),
    keywords=["wireguard"],
    license_file="./LICENSE",
    platforms="Linux",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 3 - Alpha", "Environment :: Console"
    ],
    python_requires='>=3.6',
)



import unittest
from .tables import Table, ConnectionTable


class TestTable(unittest.TestCase):
  def test_creation(self):
    #Test table creation
    Table(3, 4)
    Table(3, 4, row_names=["Alpha", "Beta", "Gamma"])
    Table(3, 4, column_names=["Alpha", "Beta", "Gamma", "Delta"])
    t = Table(3, 4, ["Alpha", "Beta", "Gamma"], ["Alpha", "Beta", "Gamma", "Delta"])

    #Test initialization values
    for i in range(3):
      for j in range(4):
        self.assertEqual(None, t.content[i][j])

  def test_getter_and_setter(self):
    t = Table(3, 4, ["Alpha", "Beta", "Gamma"], ["Alpha", "Beta", "Gamma", "Delta"])

    r = [1, 2, 3, 4]
    for i in range(3):
      for j in range(4):
        self.assertEqual(None, t.getitem(i, j))
        t.setitem(i, j, "a")
        self.assertEqual("a", t.getitem(i, j))
      t.setrow(i, r)
      for j in range(4):
        self.assertEqual(r[j], t.getitem(i, j))


  def test_str(self):
    t = Table(3, 4, ["Alpha", "Beta", "Gamma"], ["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "      Alpha Ypsilon Beta Delta\nAlpha None  None    None None \n Beta None  None    None None \nGamma None  None    None None "
    self.assertEqual(s, str(t))

  def test_repr(self):
    t = Table(3, 4, ["Alpha", "Beta", "Gamma"], ["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "Table(3, 4, ['Alpha', 'Beta', 'Gamma'], ['Alpha', 'Ypsilon', 'Beta', 'Delta'])"
    self.assertEqual(s, repr(t))

  def test_error(self):
    self.assertRaises(ValueError, Table, 3, 4, ["Alpha"])
    self.assertRaises(ValueError, Table, 3, 4, [], ["Alpha"])
    self.assertRaises(ValueError, Table ,3, 4, ["Alpha"], ["Alpha"])


class TestConnectionTable(unittest.TestCase):
  def test_creation(self):
    #Test table creation
    t = ConnectionTable( ["Alpha", "Beta", "Gamma", "Delta"])

    #Test initialization values
    for i in range(4):
      for j in range(4):
        self.assertEqual(0, t.content[i][j])

    for i in range(4):
      self.assertEqual(None, t.content[i][4])

  def test_getter_and_setter(self):
    t = ConnectionTable( ["Alpha", "Beta", "Gamma", "Delta"])

    r = [1, 2, 3, 4, "Alpha"]
    for i in range(4):
      for j in range(4):
        self.assertEqual(0, t.getitem(i, j))
        if (i == j):
          pass
        else:
          t.setitem(i, j, 1)
          self.assertEqual(1, t.getitem(i, j))
      t.setrow(i, r)
      for j in range(5):
        self.assertEqual(r[j], t.getitem(i, j))

  def test_str(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     0       0    0     None     \nYpsilon 0     0       0    0     None     \n   Beta 0     0       0    0     None     \n  Delta 0     0       0    0     None     "
    self.assertEqual(s, str(t))

  def test_repr(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "ConnectionTable(['Alpha', 'Ypsilon', 'Beta', 'Delta'])"
    self.assertEqual(s, repr(t))

  def test_update(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     0       0    0     None     \nYpsilon 0     0       0    0     None     \n   Beta 0     0       0    0     None     \n  Delta 0     0       0    0     None     "
    self.assertEqual(s, str(t))

    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     0       0    0     None     \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Delta    \n  Delta 1     1       1    0     Beta     "
    t.update(s)
    self.assertEqual(s, str(t))

    for i in range(t.n):
      for j in range(t.n):
        self.assertTrue(isinstance(t.getitem(i, j), int))
      if i == 0:
        self.assertTrue(t.getitem(i, t.m - 1) is None)
      else:
        self.assertTrue(isinstance(t.getitem(i, t.m - 1), str))

  def test_get_outgoing_connected_peers(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     1       1    1     Delta    \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Delta    \n  Delta 1     1       1    0     Beta     "
    t.update(s)

    l = t.get_outgoing_connected_peers("Alpha")
    self.assertEqual(["Ypsilon", "Beta", "Delta"], l)

  def test_get_ingoing_connected_peers(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     1       1    1     Delta    \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Delta    \n  Delta 1     1       1    0     Beta     "
    t.update(s)


    l = t.get_ingoing_connected_peers("Ypsilon")
    self.assertEqual(["Alpha", "Beta", "Delta"], l)


  def test_get_main_peer(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     1       1    1     Delta    \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Delta    \n  Delta 1     1       1    0     Beta     "
    t.update(s)

    self.assertEqual("Delta", t.get_main_peer("Alpha"))



  def test_error(self):
    t = ConnectionTable(["Alpha", "Ypsilon", "Beta", "Delta"])
    self.assertRaises(ValueError, t.setitem, 1, 1, 1)
    self.assertRaises(ValueError, t.setitem, 1, 4, 1)


    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 0     1       1    1     Delta    \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Beta     \n  Delta 1     1       1    0     Beta     "
    self.assertRaises(ValueError, t.update, s)


    s = "        Alpha Ypsilon Beta Delta main_peer\n  Alpha 1     1       1    1     Delta    \nYpsilon 1     0       1    1     Delta    \n   Beta 1     1       0    1     Delta    \n  Delta 1     1       1    0     Beta     "
    self.assertRaises(ValueError, t.update, s)


if __name__ == "__main__":
  unittest.main()



from .console import print_error
from .console import print_message

from .peers import add_peer
from .peers import delete_peer
from .peers import edit_peer_connections
from .peers import list_peers
from .peers import rekey_peer

from .shared import create_wireguard_config
from .shared import yes_no_menu

from .sites import add_site
from .sites import delete_site
from .sites import get_site_name
from .sites import list_sites

from ..library import WireUI


def start_message():
  print_message(0, "wireui")
  print_message(0, "Tool for creating and managing wireguard configs")
  print_message(0, "Version 0.1.0a")
  print_message(0, "(C) 2020 Tim Schlottmann")
  print_message(0, "")


def entrypoint_menu(w: WireUI):
  while True:
    sites_count = list_sites(w)
    # if sites_count:
    #   print_message(0, "0: List sites")
    print_message(0, "1: Create new site")
    if sites_count:
      print_message(0, "2: Edit existing site")
      print_message(0, "3: Delete existing site")
      print_message(0, "4: Recreate config files")
    print_message(0, "q: Exit")
    choice = input("What do you want to do? ")

    # if choice == "0" and sites_count:
    #   list_sites(w)
    if choice == "1":
      site_menu(w, add_site(w))
    elif choice == "2" and sites_count:
      list_sites(w)
      site_menu(w, get_site_name(w, should_exist=True))
    elif choice == "3" and sites_count:
      list_sites(w)
      delete_site(w)
    elif choice == "4" and sites_count:
      list_sites(w)
      create_wireguard_config(w, get_site_name(w, should_exist=True))
    elif choice == "q":
      _exit(0)

    print_message(0, "")


def site_menu(w: WireUI, site_name: str):
  leave = False
  while not leave:
    peers_count = list_peers(w, site_name)

    # TODO: Remove this hint when editing is possible
    print_message(0, "For full editing of peers please use the sites.json file!")

    # if peers_count:
    #   print_message(0, "0: List peers")
    print_message(0, "1: Add peer")
    if peers_count:
      print_message(0, "2: Create new keys for a peer")
      print_message(0, "3: Delete peer")
      print_message(0, "4: Edit connection table")
    print_message(0, "b: Back")
    print_message(0, "q: Exit")
    choice = input("What do you want to do? ")

    # if choice == "0" and peers_count:
    #   list_peers(w, site_name)
    if choice == "1":
      add_peer(w, site_name)
    elif choice == "2" and peers_count:
      list_peers(w, site_name)
      rekey_peer(w, site_name)
    elif choice == "3" and peers_count:
      list_peers(w, site_name)
      delete_peer(w, site_name)
    elif choice == "4" and peers_count:
      edit_peer_connections(w, site_name)
    elif choice == "b":
      leave = True
    elif choice == "q":
      _exit(0)

    print_message(0, "")


def _exit(i):
  print_message(0, "Bye")
  exit(i)



from .dicts import JSONDecodeError
from .dicts import JsonDict

from .exceptions import DataIntegrityError
from .exceptions import KeyDoesExistError
from .exceptions import KeyDoesNotExistError
from .exceptions import PeerDoesExistError
from .exceptions import PeerDoesNotExistError
from .exceptions import SettingDoesExistError
from .exceptions import SettingDoesNotExistError
from .exceptions import SiteDoesExistError
from .exceptions import SiteDoesNotExistError
from .exceptions import WireguardNotFoundError

from .peers import Keys
from .peers import PeerItems
from .peers import Peers

from .settings import Settings

from .sites import SiteItems
from .sites import Sites

from .tables import ConnectionTable

__all__ = [
    "ConnectionTable",
    "DataIntegrityError",
    "JSONDecodeError",
    "JsonDict",
    "Keys",
    "KeyDoesExistError",
    "KeyDoesNotExistError",
    "PeerItems",
    "PeerDoesExistError",
    "PeerDoesNotExistError",
    "Peers",
    "SettingDoesExistError",
    "SettingDoesNotExistError",
    "Settings",
    "SiteItems",
    "SiteDoesExistError",
    "SiteDoesNotExistError",
    "Sites",
    "WireguardNotFoundError",
]



# peers.py
# Peers for wireguard
# Author: Tim Schlottmann

# from collections import UserDict  # creates not JSON serializable error
from typing import Dict

# TODO: Evaluate TypedDict when python3.8 is available
Keys = dict

# TODO: Evaluate TypedDict when python3.8 is available
PeerItems = dict


class Peers(dict):
  """ Peers for wireguard """

  def __getitem__(self, peer_name) -> PeerItems:
    return super().__getitem__(peer_name)

  def __setitem__(self, peer_name, peer: PeerItems):
    super().__setitem__(peer_name, peer)
